#include <iostream>
#include <string>
#include <queue>
#include <map>
#include "graph.h"
#include "argparser.h"

using namespace std;

void printUsage(char* progname);

void ecmp(Graph * graphp);
void lfa(Graph * graphp);
void uturns(Graph * graphp);
void tunnels(Graph * graphp, int levels, bool df);
void rule2(Graph * graphp);
void rule3(Graph * graphp);
bool isLinkInPath(int a, int b, vector<int> path);
void outputCDF(map<int,int> & pdf, ostream & os);

int main(int argc, char **argv) {
	string input;
	string mechanism;
	

	ArgParser argp;
	try {
		argp.addSpec("i", true);
		argp.addSpec("m", true);
		argp.addSpec("l", true);
		argp.addSpec("d", true);
		argp.parse(argc, argv);
		
		input = argp.getNonEmptyString("i");
		mechanism = argp.getNonEmptyString("m");
		
	} catch (invalid_argument e) {
		cerr << "std::invalid_argument exception: " << e.what() << endl;
		printUsage(argv[0]);
		exit(1);
	}
	
	Graph* graphp;
	
	graphp = new Graph(input.c_str());
		
	if (mechanism == "ecmp"){
		ecmp(graphp);
	} else if (mechanism == "lfa") {
		lfa(graphp);
	} else if (mechanism == "uturns") {
		uturns(graphp);
	} else if (mechanism == "tunnels") {
		int levels = 0;
		bool df = false;
		try {
			
			levels = argp.getInt("l");
			df = argp.getBool("d");
			
		} catch (invalid_argument e) {
			cerr << "std::invalid_argument exception: " << e.what() << endl;
			cerr << "either -l or -d parameter not set";
			exit(1);
		}
		tunnels(graphp, levels, df);
	} else if (mechanism == "rule2") {
		rule2(graphp);
	}
	
	/*Node * src = graphp->nodes[0];
	Node * dst = graphp->nodes[10];
	
	queue<int> q;
	map<int,int> queued;
	map<int,bool> visited;
	map<int, vector<int> > mapping;
	
	q.push(src->getId());
	bool finish=false;
	int level = 0;
	while (!q.empty() && !finish) {
		Node * node = graphp->nodes[q.front()];
		q.pop();
		visited[node->getId()]=true;
		cout << "Visiting node " << node->getId() << " Level " << queued[node->getId()] <<endl;
		vector<Node *> nodesList = node->getNeighbors();
		for (vector<Node*>::iterator iter = nodesList.begin(); iter != nodesList.end(); ++iter) { 
				level++;
				Node * nd = * iter;
				int ndId = nd->getId();
				if ( (visited.find(ndId) == visited.end() ) && ( queued.find(ndId) == queued.end() )  ) {
					cout << "Adding in queue node  " << ndId << endl;
					q.push(ndId);
					int hopCount = src->routingTable[ndId].hopCount;
					queued[ndId] = hopCount;
					//if ( mapping.find( hopCount ) == mapping.end() ) mapping[hopCount]=
					mapping[hopCount].push_back(ndId);
				}
		}
	}
	cout << "---------" <<endl ;
	for (map<int, vector<int> >::iterator iter = mapping.begin(); iter != mapping.end(); ++iter) { 
		int level = iter->first;
		vector<int> list = iter->second;
		cout << "Level "  << level << "   " ;
		for (vector<int>::iterator iterlist = list.begin(); iterlist != list.end(); ++iterlist) { 
			cout <<  *iterlist << "  ";
		}
		cout << endl;
		
	}*/
	
	return 0;
}

void ecmp(Graph * graphp){
	int ecmpLinkSuccess = 0;
	int ecmpLinkNeeded = 0;
	int ecmpNodeSuccess = 0;
	int ecmpNodeNeeded = 0;
	
	map<int,int> freqLinksNeighb;
	map<int,int> freqNodesNeighb;
	
	//consider each node as a source
	for (vector<Node*>::iterator iter = graphp->nodes.begin(); iter != graphp->nodes.end(); ++iter) {
		Node * src = *iter;
		int srcId = src->getId();
		
		//find neighbors for chosen source
		vector<Node *> nbrsList = src->getNeighbors();
		for (vector<Node*>::iterator iter = nbrsList.begin(); iter != nbrsList.end(); ++iter) {
			Node * nbr = *iter;
			//cout << nbr->getId() << endl;
		}
		
		//consider each node as a destination
		for (vector<Node*>::iterator iter = graphp->nodes.begin(); iter != graphp->nodes.end(); ++iter) {
			Node* dst = *iter;
			int  dstId = dst->getId();
			graphp->recordSPPerPair(src, dstId);
			vector<int> path = src->routingTable[dstId].shortest.nids; //path from src to destination
			
			if  ( dstId != src->getId() ) { //count only src!=destination
				for (vector<Node*>::iterator iter = nbrsList.begin(); iter != nbrsList.end(); ++iter) {
					//for every neighbor (assume neighbor link/node is dead)
					Node * nbr = *iter;

					
					//************LINK FAILURES************	
					bool hasLinkECMP = false;
					//only compute ECMPs for cases where the next hop in the shortest path
					//contains chosen failed link
					if (  ( src->getNextHopToDst(dstId) == nbr->getId() )   )  {
						ecmpLinkNeeded++;
						int numOfLinksPerPair = 0;
						//get all other neighbors (failed neighbor will be exluded inside)
						for (vector<Node*>::iterator myiter = nbrsList.begin(); myiter != nbrsList.end(); ++myiter) { 
							Node * mynbr = *myiter;
							
							//avoid checking for alternate path using the link connected 
							//to the assumed failed neighbouring router
							if ( mynbr->getId() == nbr->getId() ) continue; 
						
							if ( mynbr->getCostToDst(dstId) == src->getCostToDst(dstId) - graphp->getLinkCost(srcId,mynbr->getId() ) ) {
					
								//does this ECMP provide a loop free path?
								//check if neigbor's next hop to the destination is the source itself
								if (mynbr->getNextHopToDst(dstId) != srcId) {
									//cout << "+++" ;
									hasLinkECMP= true;
									numOfLinksPerPair++ ;
								} //else cout << "???" ;
								
							}
							//else cout << "---" ; 
								
							/*cout << " For Dst \t" << dstId << 
								" to avoid LINK \t" << srcId << "-" << nbr->getId() <<
								"\t using neighbor \t" << mynbr->getId() <<
								"\t cost(n[i+1]) \t" << mynbr->getCostToDst(dstId) <<
								"\t cost(n[i]) \t" << src->getCostToDst(dstId) <<
								"\t link(n[i],n[i+1]) \t" << graphp->getLinkCost(srcId,mynbr->getId() ) << endl;*/
						}
						if (hasLinkECMP) ecmpLinkSuccess++;
						freqLinksNeighb[numOfLinksPerPair]++;
					}
						
						
					//************NODE FAILURES************
					bool hasNodeECMP = false;
					//only compute ECMPs for cases where chosen failed node
					//is in the shortest path towards the destination
					
					//is failed neighbor in the path from src to dst?
					vector<int>::iterator result = std::find(path.begin(), path.end(), nbr->getId() );
					if ( result != path.end() )  {
						ecmpNodeNeeded++;
						int numOfNodesPerPair = 0 ;
						//get all other neighbors (failed neighbor will be exluded inside)
						for (vector<Node*>::iterator myiter = nbrsList.begin(); myiter != nbrsList.end(); ++myiter) { 
							Node * mynbr = *myiter;

							//avoid checking for alternate path using the link connected 
							//to the assumed failed neighbouring router
							if ( mynbr->getId() == nbr->getId() ) continue; 
								
							//cout << "Failed Node resides in path :  " << *result << endl;
							if ( mynbr->getCostToDst(dstId) == src->getCostToDst(dstId) - graphp->getLinkCost(srcId,mynbr->getId() ) ) {
								
								//does this ECMP provide a loop free path?
								//check if new neigbour contains the failed node in its path
								graphp->recordSPPerPair(mynbr, dstId);
								vector<int> mypath = mynbr->routingTable[dstId].shortest.nids;
								vector<int>::iterator myresult = std::find(mypath.begin(), mypath.end(), nbr->getId() );
								
								if ( myresult == mypath.end()) {
									//cout << "+++" ;
									hasNodeECMP= true;
									numOfNodesPerPair++;
								}// else cout << "???" ;
								
							}
							//else cout << "---" ; 
								
							/*cout << " For Dst \t" << dstId << 
								" to avoid NODE \t" << srcId << "-" << nbr->getId() <<
								"\t using neighbor \t" << mynbr->getId() <<
								"\t cost(n[i+1]) \t" << mynbr->getCostToDst(dstId) <<
								"\t cost(n[i]) \t" << src->getCostToDst(dstId) <<
								"\t link(n[i],n[i+1]) \t" << graphp->getLinkCost(srcId,mynbr->getId() ) << endl;*/
						}
						if (hasNodeECMP) ecmpNodeSuccess++;
						freqNodesNeighb[numOfNodesPerPair]++;
					}
				}
			}
	    
		}
	}
	
	cout << "# Graph connectivity" << endl ; 
	graphp->outputDegree(cout);
	
	cout << "# Number of Deflection Neighbors (Link failures) | Cumulative Fraction" << endl ; 
	outputCDF(freqLinksNeighb, cout);
	cout << "Percentage >1  :" << 1-( (double)(freqLinksNeighb[1]) / ecmpLinkSuccess) << endl;

	cout << endl <<"#ECMP for link failures : " << ecmpLinkSuccess << " out of " << ecmpLinkNeeded << 
		  "  Percentage : "  <<  (double) ecmpLinkSuccess / ecmpLinkNeeded << endl;
	
	cout << endl << "********" << endl ;

	cout << "# Number of Deflection Neighbors (Node failures) | Cumulative Fraction" << endl ; 
	outputCDF(freqNodesNeighb, cout);
	cout << "Percentage >1  :" << 1-( (double)freqNodesNeighb[1]/ ecmpNodeSuccess) << endl;
	cout << "#ECMP for node failures : " << ecmpNodeSuccess << " out of " << ecmpNodeNeeded << 
		  "  Percentage : "  <<  (double) ecmpNodeSuccess / ecmpNodeNeeded << endl;
	
	
	
}

void lfa(Graph * graphp) {
	int linkSuccess = 0;
	int linkNeeded = 0;
	int nodeSuccess = 0;
	int nodeNeeded = 0;
	map<int,int> freqLinkNeighb;
	map<int,int> freqNodeNeighb;
	
	//consider each node as a source
	for (vector<Node*>::iterator iter = graphp->nodes.begin(); iter != graphp->nodes.end(); ++iter) {
		Node * src = *iter;
		int srcId = src->getId();
		
		//find neighbors for chosen source
		vector<Node *> nbrsList = src->getNeighbors();
		for (vector<Node*>::iterator iter = nbrsList.begin(); iter != nbrsList.end(); ++iter) {
			Node * nbr = *iter;
			//cout << nbr->getId() << endl;
		}
		
		//consider each node as a destination
		for (vector<Node*>::iterator iter = graphp->nodes.begin(); iter != graphp->nodes.end(); ++iter) {
			Node* dst = *iter;
			int  dstId = dst->getId();
			graphp->recordSPPerPair(src, dstId);
			vector<int> path = src->routingTable[dstId].shortest.nids; //path from src to destination
			
			if  ( dstId != srcId ) { //count only src!=destination
				for (vector<Node*>::iterator iter = nbrsList.begin(); iter != nbrsList.end(); ++iter) {
					//for every neighbor (assume neighbor link/node is dead)
					Node * nbr = *iter;
					
					//************LINK FAILURES************	
					bool hasLinkLFA = false;
					//only compute LFAs for cases where the next hop in the shortest path
					//contains chosen failed link
					if (  ( src->getNextHopToDst(dstId) == nbr->getId() )   )  {
						linkNeeded++;
						int numOfLinkPerPair = 0;
						//get all other neighbors (failed neighbor will be exluded inside)
						for (vector<Node*>::iterator myiter = nbrsList.begin(); myiter != nbrsList.end(); ++myiter) { 
							Node * mynbr = *myiter;
							
							//avoid checking for alternate path using the link connected 
							//to the assumed failed neighbouring router
							if ( mynbr->getId() == nbr->getId() ) continue; 
						
							if ( mynbr->getCostToDst(dstId) < graphp->getLinkCost(mynbr->getId(),srcId ) + src->getCostToDst(dstId) ) {
							//if ( mynbr->getCostToDst(dstId) < src->getCostToDst(dstId) )  {
								
								//does this LFA provide a loop free path?
								//check if neigbor's next hop to the destination is the source itself
								if (mynbr->getNextHopToDst(dstId) != srcId) {
									//cout << "+++" ;
									hasLinkLFA= true;
									numOfLinkPerPair++;
								} //else cout << "???" ;
								
							}
							//else cout << "---" ; 
								
							//~ cout << " For Dst \t" << dstId << 
								//~ " to avoid LINK \t" << srcId << "-" << nbr->getId() <<
								//~ "\t using neighbor \t" << mynbr->getId() <<
								//~ "\t cost(n[i+1]) \t" << mynbr->getCostToDst(dstId) <<
								//~ "\t cost(n[i]) \t" << src->getCostToDst(dstId) <<
								//~ "\t link(n[i+1],n[i]) \t" << graphp->getLinkCost(mynbr->getId(),srcId ) << endl;
						}
						freqLinkNeighb[numOfLinkPerPair]++;
						if (hasLinkLFA) linkSuccess++;
					}
						
						
					//************NODE FAILURES************
					bool hasNodeLFA = false;
					//only compute LFAs for cases where chosen failed node
					//is in the shortest path towards the destination
					
					//is failed neighbor in the path from src to dst?
					vector<int>::iterator result = std::find(path.begin(), path.end(), nbr->getId() );
					if ( result != path.end() )  {
						nodeNeeded++;
						int numOfNodePerPair = 0;
						//get all other neighbors (failed neighbor will be exluded inside)
						for (vector<Node*>::iterator myiter = nbrsList.begin(); myiter != nbrsList.end(); ++myiter) { 
							Node * mynbr = *myiter;

							//avoid checking for alternate path using the link connected 
							//to the assumed failed neighbouring router
							if ( mynbr->getId() == nbr->getId() ) continue; 
								
							//cout << "Failed Node resides in path :  " << *result << endl;
							if ( mynbr->getCostToDst(dstId) < graphp->getLinkCost(mynbr->getId(),srcId ) + src->getCostToDst(dstId) ) {
							//if ( mynbr->getCostToDst(dstId) < src->getCostToDst(dstId) )  {
								
								//does this LFA provide a loop free path?
								//check if new neigbour contains the failed node in its path
								graphp->recordSPPerPair(mynbr, dstId);
								vector<int> mypath = mynbr->routingTable[dstId].shortest.nids;
								vector<int>::iterator myresult = std::find(mypath.begin(), mypath.end(), nbr->getId() );
								
								if ( myresult == mypath.end()) {
									//cout << "+++" ;
									hasNodeLFA= true;
									numOfNodePerPair++;
								} //else cout << "???" ;
								
							}
							//else cout << "---" ; 
								
								//~ cout << " For Dst \t" << dstId << 
								//~ " to avoid NODE \t" << srcId << "-" << nbr->getId() <<
								//~ "\t using neighbor \t" << mynbr->getId() <<
								//~ "\t cost(n[i+1]) \t" << mynbr->getCostToDst(dstId) <<
								//~ "\t cost(n[i]) \t" << src->getCostToDst(dstId) <<
								//~ "\t link(n[i+1],n[i]) \t" << graphp->getLinkCost(mynbr->getId(),srcId ) << endl;
						}
						if (hasNodeLFA) nodeSuccess++;
						freqNodeNeighb[numOfNodePerPair]++;
					}
				}
			}
	    
		}
	}
	
	cout << "# Graph connectivity" << endl ; 
	graphp->outputDegree(cout);
	
	cout << "# Number of Deflection Neighbors (Link failures) | Cumulative Fraction" << endl ; 
	outputCDF(freqLinkNeighb, cout);
	cout << "Percentage >1  :" << 1-( (double)freqLinkNeighb[1]/ linkSuccess) << endl;

	cout << endl <<"#LFA for link failures : " << linkSuccess << " out of " << linkNeeded << 
		  "  Percentage : "  <<  (double) linkSuccess / linkNeeded << endl;
	
	cout << endl << endl << endl ;

	cout << "# Number of Deflection Neighbors (Node failures) | Cumulative Fraction" << endl ; 
	outputCDF(freqNodeNeighb, cout);
	cout << "Percentage >1  :" << 1-( (double)freqNodeNeighb[1]/ nodeSuccess) << endl;
	cout << "#LFA for node failures : " << nodeSuccess << " out of " << nodeNeeded << 
		  "  Percentage : "  <<  (double) nodeSuccess / nodeNeeded << endl;
}

void uturns(Graph * graphp) {
	int linkSuccess = 0;
	int linkNeeded = 0;
	int nodeSuccess = 0;
	int nodeNeeded = 0;
	map<int,int> freqLinkNeighb;
	map<int,int> freqNodeNeighb;
	
	//consider each node as a source
	for (vector<Node*>::iterator iter = graphp->nodes.begin(); iter != graphp->nodes.end(); ++iter) {
		Node * src = *iter;
		int srcId = src->getId();
		
		//find neighbors for chosen source
		vector<Node *> nbrsList = src->getNeighbors();
		for (vector<Node*>::iterator iter = nbrsList.begin(); iter != nbrsList.end(); ++iter) {
			Node * nbr = *iter;
			//cout << nbr->getId() << endl;
		}
		
		//consider each node as a destination
		for (vector<Node*>::iterator iter = graphp->nodes.begin(); iter != graphp->nodes.end(); ++iter) {
			Node* dst = *iter;
			int  dstId = dst->getId();
			graphp->recordSPPerPair(src, dstId);
			vector<int> path = src->routingTable[dstId].shortest.nids; //path from src to destination
			
			if  ( dstId != srcId ) { //count only src!=destination
				for (vector<Node*>::iterator iter = nbrsList.begin(); iter != nbrsList.end(); ++iter) {
					//for every neighbor (assume neighbor link/node is dead)
					Node * nbr = *iter;
					
					//************LINK FAILURES************	
					
					//only compute UTurns for cases where the next hop in the shortest path
					//contains chosen failed link
					if (  ( src->getNextHopToDst(dstId) == nbr->getId() )   )  {
						linkNeeded++;
						int numOfLinkPerPair = 0;
						
						//get all other neighbors (failed neighbor will be exluded inside)
						for (vector<Node*>::iterator myiter = nbrsList.begin(); myiter != nbrsList.end(); ++myiter) { 
							Node * mynbr = *myiter;
							
							//avoid checking for alternate path using the link connected 
							//to the assumed failed neighbouring router
							if ( mynbr->getId() == nbr->getId() ) continue; 
							vector<int> altNodes;
							
							bool hasLinkUturns = false;
							bool hasLinkLFA = false;
							//LFAS
							if ( mynbr->getCostToDst(dstId) < graphp->getLinkCost(mynbr->getId(),srcId ) + src->getCostToDst(dstId) ) {
							
								//does this LFA provide a loop free path?
								//check if neigbor's next hop to the destination is the source itself
								if (mynbr->getNextHopToDst(dstId) != srcId) {
									hasLinkLFA= true;
									numOfLinkPerPair++;
								}
								
							}
							
							//UTurns
							if ( /*!hasLinkLFA &&*/
								mynbr->getCostToDst(dstId) >= graphp->getLinkCost(mynbr->getId(),srcId ) + src->getCostToDst(dstId) ) {
								
								//new neighbor's next hop to the destination should be the source itself
								if (mynbr->getNextHopToDst(dstId) == srcId) {
									bool hasAlternate = false;
									vector<Node *> myNbrList = mynbr->getNeighbors();
									for (vector<Node*>::iterator uIter = myNbrList.begin(); uIter != myNbrList.end(); ++uIter) {
										Node * lfa = *uIter;			
										if ( lfa->getId() == src->getId() ) continue; 
										
										//LFA condition
										if ( lfa->getCostToDst(dstId) < graphp->getLinkCost(lfa->getId(),mynbr->getId() ) + mynbr->getCostToDst(dstId) ) {
											graphp->recordSPPerPair(lfa, dstId);
											vector<int> newpath = lfa->routingTable[dstId].shortest.nids;
											
											vector<int>::iterator myresult = std::find(newpath.begin(), newpath.end(), src->getId() );
											
											
											//src should not reside in the path from lfa of Uturn neighbor to the destination
											//a more strict check could check whether the link src-nbr is in the path but we think
											//that the implemented check is enough
											if ( myresult == newpath.end()) { 
											//if (!isLinkInPath(srcId,nbr->getId(), newpath) ) { //strict implementation (but slower)
												hasAlternate = true;
												altNodes.push_back(lfa->getId());
											}
										}											
									}
									
									if (hasAlternate) {
										hasLinkUturns = true;
										numOfLinkPerPair++;
									}
									
								} 
								
							}
								
							//~ cout << " For Dst \t" << dstId << 
								//~ " to avoid LINK \t" << srcId << "-" << nbr->getId() <<
								//~ "\t using UTURN neighbor \t" << mynbr->getId() <<
								//~ "\t cost(n[i+1]) \t" << mynbr->getCostToDst(dstId) <<
								//~ "\t cost(n[i]) \t" << src->getCostToDst(dstId) <<
								//~ "\t link(n[i+1],n[i]) \t" << graphp->getLinkCost(mynbr->getId(),srcId ) << "\t" ;
							//~ for (vector<int>::iterator lfaIter = altNodes.begin(); lfaIter != altNodes.end(); ++ lfaIter) {
								//~ cout << *lfaIter << ", ";
							//~ }
							//~ cout <<endl;
							
						}
						freqLinkNeighb[numOfLinkPerPair]++;
						if (numOfLinkPerPair ) linkSuccess++;
					}
						
						
					//************NODE FAILURES************
					
					//only compute UTurns for cases where chosen failed node
					//is in the shortest path towards the destination
					
					//is failed neighbor in the path from src to dst?
					vector<int>::iterator result = std::find(path.begin(), path.end(), nbr->getId() );
					if ( result != path.end() )  {
						
						nodeNeeded++;
						int numOfNodePerPair = 0;
						
						//get all other neighbors (failed neighbor will be exluded inside)
						for (vector<Node*>::iterator myiter = nbrsList.begin(); myiter != nbrsList.end(); ++myiter) { 
							Node * mynbr = *myiter;

							//avoid checking for alternate path using the assumed failed neighbouring router
							if ( mynbr->getId() == nbr->getId() ) continue; 
							
							vector<int> altNodes;
							bool hasNodeUturns = false;
							bool hasNodeLFA= false;
							
							//LFAS
							if ( mynbr->getCostToDst(dstId) < graphp->getLinkCost(mynbr->getId(),srcId ) + src->getCostToDst(dstId) ) {
														
								//does this LFA provide a loop free path?
								//check if new neigbour contains the failed node in its path
								graphp->recordSPPerPair(mynbr, dstId);
								vector<int> mypath = mynbr->routingTable[dstId].shortest.nids;
								vector<int>::iterator myresult = std::find(mypath.begin(), mypath.end(), nbr->getId() );
								
								if ( myresult == mypath.end()) {
									hasNodeLFA= true;
									numOfNodePerPair++;
								}
							}
							
							//UTurns
							if (/*!hasNodeLFA &&*/
								mynbr->getCostToDst(dstId) >= graphp->getLinkCost(mynbr->getId(),srcId ) + src->getCostToDst(dstId) ) {
								
								//new neighbor's next hop to the destination should be the source itself
								if (mynbr->getNextHopToDst(dstId) == srcId) {
									bool hasAlternate = false;
									vector<Node *> myNbrList = mynbr->getNeighbors();
									for (vector<Node*>::iterator uIter = myNbrList.begin(); uIter != myNbrList.end(); ++uIter) {
										Node * lfa = *uIter;			
										if ( lfa->getId() == src->getId() ||
											 lfa->getId() == nbr->getId() ) continue; 
										
										//LFA condition
										if ( lfa->getCostToDst(dstId) < graphp->getLinkCost(lfa->getId(),mynbr->getId() ) + mynbr->getCostToDst(dstId) ) {
											graphp->recordSPPerPair(lfa, dstId);
											vector<int> newpath = lfa->routingTable[dstId].shortest.nids;
											
											vector<int>::iterator myresult = std::find(newpath.begin(), newpath.end(), nbr->getId() );
											
											
											//failed neighbor "nbr" should not reside in the path from the lfa of 
											//Uturn neighbor to the destination
											if ( myresult == newpath.end()) { 
												hasAlternate = true;
												altNodes.push_back(lfa->getId());
											}
										}											
									}
									
									if (hasAlternate) {
										hasNodeUturns = true;
										numOfNodePerPair++;
									}
									
								}
							}								
							
							//~ cout << " For Dst \t" << dstId << 
							//~ " to avoid NODE \t" << srcId << "-" << nbr->getId() <<
							//~ "\t using neighbor \t" << mynbr->getId() <<
							//~ "\t cost(n[i+1]) \t" << mynbr->getCostToDst(dstId) <<
							//~ "\t cost(n[i]) \t" << src->getCostToDst(dstId) <<
							//~ "\t link(n[i+1],n[i]) \t" << graphp->getLinkCost(mynbr->getId(),srcId ) << endl;
						}
						if (numOfNodePerPair) nodeSuccess++;
						freqNodeNeighb[numOfNodePerPair]++;
					}
				}
			}
	    
		}
	}
	
	//cout << "# Graph connectivity" << endl ; 
	//graphp->outputDegree(cout);
	
	cout << "# Number of Deflection Neighbors (Link failures) | Cumulative Fraction" << endl ; 
	outputCDF(freqLinkNeighb, cout);
	cout << "Percentage >1  :" << 1-( (double)freqLinkNeighb[1]/ linkSuccess) << endl;

	cout << endl <<"#UTurn for link failures : " << linkSuccess << " out of " << linkNeeded << 
		  "  Percentage : "  <<  (double) linkSuccess / linkNeeded << endl;
	
	cout << endl << endl << endl ;

	cout << "# Number of Deflection Neighbors (Node failures) | Cumulative Fraction" << endl ; 
	outputCDF(freqNodeNeighb, cout);
	cout << "Percentage >1  :" << 1-( (double)freqNodeNeighb[1]/ nodeSuccess) << endl;
	cout << "#Uturn for node failures : " << nodeSuccess << " out of " << nodeNeeded << 
		  "  Percentage : "  <<  (double) nodeSuccess / nodeNeeded << endl;
	
}

void tunnels(Graph * graphp, int levels, bool df){
	int linkSuccess = 0;
	int linkNeeded = 0;
	int nodeSuccess = 0;
	int nodeNeeded = 0;
	map<int,int> freqLinkNeighb;
	map<int,int> freqNodeNeighb;
	
	//consider each node as a source
	for (vector<Node*>::iterator iter = graphp->nodes.begin(); iter != graphp->nodes.end(); ++iter) {
		Node * src = *iter;
		int srcId = src->getId();


		queue<int> q;
		map<int,int> queued;
		map<int,bool> visited;
		map<int, vector<int> > mapping;
		
		//BFS computation
		q.push(src->getId());
		bool finish=false;
		int level = 0;
		while (!q.empty() && !finish) {
			Node * node = graphp->nodes[q.front()];
			q.pop();
			visited[node->getId()]=true;
			//cout << "Visiting node " << node->getId() << " Level " << queued[node->getId()] <<endl;
			vector<Node *> nodesList = node->getNeighbors();
			for (vector<Node*>::iterator iter = nodesList.begin(); iter != nodesList.end(); ++iter) { 
					level++;
					Node * nd = * iter;
					int ndId = nd->getId();
					if ( (visited.find(ndId) == visited.end() ) && ( queued.find(ndId) == queued.end() )  ) {
						//cout << "Adding in queue node  " << ndId << endl;
						q.push(ndId);
						int hopCount = src->routingTable[ndId].hopCount;
						queued[ndId] = hopCount;
						mapping[hopCount].push_back(ndId);
					}
			}
		}
		
		//find neighbors for chosen source
		vector<Node *> nbrsList = src->getNeighbors();
		for (vector<Node*>::iterator iter = nbrsList.begin(); iter != nbrsList.end(); ++iter) {
			Node * nbr = *iter;
			//cout << nbr->getId() << endl;
		}
		
		//consider each node as a destination
		for (vector<Node*>::iterator iter = graphp->nodes.begin(); iter != graphp->nodes.end(); ++iter) {
			Node* dst = *iter;
			int  dstId = dst->getId();
			graphp->recordSPPerPair(src, dstId);
			vector<int> path = src->routingTable[dstId].shortest.nids; //path from src to destination
			
			if  ( dstId != srcId ) { //count only src!=destination
				for (vector<Node*>::iterator iter = nbrsList.begin(); iter != nbrsList.end(); ++iter) {
					//for every neighbor (assume neighbor link/node is dead)
					Node * nbr = *iter;
					
					//************LINK FAILURES************	
					
					//only compute UTurns for cases where the next hop in the shortest path
					//contains chosen failed link
					if (  ( src->getNextHopToDst(dstId) == nbr->getId() )   )  {
						linkNeeded++;
						int numOfLinkPerPair = 0;
						bool hasLinkTunnels = false;
						//get all other neighbors (failed neighbor will be exluded inside)
						
						for (int level=1;level<=levels;level++) {
							vector<int> list = mapping[level];
							for (vector<int>::iterator iterlist = list.begin(); iterlist != list.end(); ++iterlist) { 
								int ndId = *iterlist;
								
								//avoid checking for alternate path using the link connected 
								//to the assumed failed neighbouring router
								if ( ndId == nbr->getId() ) continue; 
									
								Node * nd = graphp->nodes[ndId];
								graphp->recordSPPerPair(src, ndId);
								
								vector<int> tunnelFromSrc = src->routingTable[ndId].shortest.nids;
								
								//failed link is src-nbr
								if (!isLinkInPath(srcId,nbr->getId(), tunnelFromSrc) ) {
									graphp->recordSPPerPair(nd, dstId);
								
									vector<int> tunnelToDst = nd->routingTable[dstId].shortest.nids;	
									if (!isLinkInPath(srcId,nbr->getId(), tunnelToDst) ) {
										hasLinkTunnels = true;
										numOfLinkPerPair++;
									}
								}
								
							}							
							
						}
						freqLinkNeighb[numOfLinkPerPair]++;
						if (hasLinkTunnels ) linkSuccess++;
						
							//~ cout << " For Dst \t" << dstId << 
								//~ " to avoid LINK \t" << srcId << "-" << nbr->getId() <<
								//~ "\t using UTURN neighbor \t" << mynbr->getId() <<
								//~ "\t cost(n[i+1]) \t" << mynbr->getCostToDst(dstId) <<
								//~ "\t cost(n[i]) \t" << src->getCostToDst(dstId) <<
								//~ "\t link(n[i+1],n[i]) \t" << graphp->getLinkCost(mynbr->getId(),srcId ) << "\t" ;
							//~ for (vector<int>::iterator lfaIter = a.begin(); lfaIter != a.end(); ++ lfaIter) {
								//~ cout << *lfaIter << ", ";
							//~ }
							//~ cout <<endl;						
						
					}
						
						
					//************NODE FAILURES************
					
					//only compute Tunnels for cases where chosen failed node
					//is in the shortest path towards the destination
					//is failed neighbor in the path from src to dst?
					vector<int>::iterator result = std::find(path.begin(), path.end(), nbr->getId() );
					if ( result != path.end() )  {
					
						nodeNeeded++;
						int numOfNodePerPair = 0;
						bool hasNodeTunnels = false;
						
						for (int level=1;level<=levels;level++) {
							vector<int> list = mapping[level];
							for (vector<int>::iterator iterlist = list.begin(); iterlist != list.end(); ++iterlist) { 
								int ndId = *iterlist;
								
								//avoid checking for alternate path using the assumed failed neighbouring router
								if ( ndId == nbr->getId() ) continue; 
								
								Node * nd = graphp->nodes[ndId];
								graphp->recordSPPerPair(src, ndId);
								
								vector<int> tunnelFromSrc = src->routingTable[ndId].shortest.nids;
								
								vector<int>::iterator resultFromSrc  = std::find(tunnelFromSrc.begin(), tunnelFromSrc.end(), nbr->getId() );
								
								if ( resultFromSrc == tunnelFromSrc.end()) {
									graphp->recordSPPerPair(nd, dstId);
								
									vector<int> tunnelToDst = nd->routingTable[dstId].shortest.nids;	
									
									vector<int>::iterator resultToDst  = std::find(tunnelToDst.begin(), tunnelToDst.end(), nbr->getId() );
								
									if ( resultToDst == tunnelToDst.end()) {
										hasNodeTunnels = true;
										numOfNodePerPair++;
									}
								}
								
							}							
							
						}
						if (hasNodeTunnels) nodeSuccess++;
						freqNodeNeighb[numOfNodePerPair]++;
						
					}
				}
			}
	    
		}
	}
	
	//cout << "# Graph connectivity" << endl ; 
	//graphp->outputDegree(cout);
	
	cout << "# Number of Deflection Neighbors (Link failures) | Cumulative Fraction" << endl ; 
	outputCDF(freqLinkNeighb, cout);
	cout << "Percentage >1  :" << 1-( (double)freqLinkNeighb[1]/ linkSuccess) << endl;

	cout << endl <<"#Tunnels for link failures : " << linkSuccess << " out of " << linkNeeded << 
		  "  Percentage : "  <<  (double) linkSuccess / linkNeeded << endl;
	
	cout << endl << endl << endl ;

	cout << "# Number of Deflection Neighbors (Node failures) | Cumulative Fraction" << endl ; 
	outputCDF(freqNodeNeighb, cout);
	cout << "Percentage >1  :" << 1-( (double)freqNodeNeighb[1]/ nodeSuccess) << endl;
	cout << "#Tunnels for node failures : " << nodeSuccess << " out of " << nodeNeeded << 
		  "  Percentage : "  <<  (double) nodeSuccess / nodeNeeded << endl;
	
	
}

void rule2(Graph * graphp){
	double linkSuccess = 0;
	int linkNeeded = 0;
	double nodeSuccess = 0;
	int nodeNeeded = 0;
	
	//consider each node as a source
	for (vector<Node*>::iterator iter = graphp->nodes.begin(); iter != graphp->nodes.end(); ++iter) {
		Node * src = *iter;
		int srcId = src->getId();
		
		//find neighbors for chosen source
		vector<Node *> nbrsList = src->getNeighbors();
		for (vector<Node*>::iterator iter = nbrsList.begin(); iter != nbrsList.end(); ++iter) {
			Node * nbr = *iter;
			//cout << nbr->getId() << endl;
		}
		
		//consider each node as a destination
		for (vector<Node*>::iterator iter = graphp->nodes.begin(); iter != graphp->nodes.end(); ++iter) {
			Node* dst = *iter;
			int  dstId = dst->getId();
			graphp->recordSPPerPair(src, dstId);
			vector<int> path = src->routingTable[dstId].shortest.nids; //path from src to destination
			
			if  ( dstId != srcId ) { //count only src!=destination
				for (vector<Node*>::iterator iter = nbrsList.begin(); iter != nbrsList.end(); ++iter) {
					//for every neighbor (assume neighbor link/node is dead)
					Node * nbr = *iter;
					
					//************LINK FAILURES************	
					
					//only compute LFAs for cases where the next hop in the shortest path
					//contains chosen failed link
					if (  ( src->getNextHopToDst(dstId) == nbr->getId() )   )  {
						linkNeeded++;
					
						double linkSuccessSubsetRuleTwo = 0;
						int neighborsSubsetRuleTwoApplies = 0;
						bool hasLinkStrictLFA = false;
						
						
						//get all other neighbors (failed neighbor will be exluded inside)
						for (vector<Node*>::iterator myiter = nbrsList.begin(); myiter != nbrsList.end(); ++myiter) { 
							Node * mynbr = *myiter;
							
							//avoid checking for alternate path using the link connected 
							//to the assumed failed neighbouring router
							if ( mynbr->getId() == nbr->getId()) continue; 
							
							//LFAS
							if ( mynbr->getCostToDst(dstId)  < src->getCostToDst(dstId) ) {
								//does this LFA provide a loop free path?
								//check if neigbor's next hop to the destination is the source itself
								if (mynbr->getNextHopToDst(dstId) != srcId) {
									hasLinkStrictLFA = true;
								}
								
							}
							
							if (!hasLinkStrictLFA) {
								neighborsSubsetRuleTwoApplies++;
								int prevNbrsNeeded = 0;
								int prevNbrsSuccess = 0;
							
								for (vector<Node*>::iterator previter = nbrsList.begin(); previter != nbrsList.end(); ++previter) { 
									Node * prevnbr = *previter;
									if ( prevnbr->getId() == nbr->getId() ||     //prev neighbor = failed neighbor
										 prevnbr->getNextHopToDst(dstId) != srcId ||  //prev neighbor does not have SRC as the next hop to DST
										 prevnbr->getId() == mynbr->getId()) continue; //prev neighbor = recovery neighbor
						
									prevNbrsNeeded++;
									
									//rule 2 condition (half checked. the other half checked above)
									if ( mynbr->getCostToDst(dstId) < prevnbr->getCostToDst(dstId) )  {
								
										//does this alternate neighbor provide a loop free path?
										//check if neigbor's next hop to the destination is the source itself
										if (mynbr->getNextHopToDst(dstId) != srcId) {
											//cout << "+++" ;
											prevNbrsSuccess++;
										} //else cout << "???" ;
								
									}
									//else cout << "---" ; 
								
									//~ cout << " For Dst \t" << dstId << 
										//~ " with prevNeighb \t" <<  prevnbr->getId() << 
										//~ " to avoid LINK \t" << srcId << "-" << nbr->getId() <<
										//~ "\t using neighbor \t" << mynbr->getId() <<
										//~ "\t cost(n[i+1]) \t" << mynbr->getCostToDst(dstId) <<
										//~ "\t cost(n[i]) \t" << src->getCostToDst(dstId) << endl;
								}
								if (prevNbrsNeeded) {
									linkSuccessSubsetRuleTwo += (double) prevNbrsSuccess/ prevNbrsNeeded ;
								}		
							}	
						}
						

						if (hasLinkStrictLFA) 
							linkSuccess++;
						else if (linkSuccessSubsetRuleTwo!=0 && neighborsSubsetRuleTwoApplies!=0)
							linkSuccess += linkSuccessSubsetRuleTwo/neighborsSubsetRuleTwoApplies;
					}
						
						
						
					//************NODE FAILURES************
					
					//only compute LFAs for cases where chosen failed node
					//is in the shortest path towards the destination
					
					//is failed neighbor in the path from src to dst?
					vector<int>::iterator result = std::find(path.begin(), path.end(), nbr->getId() );
					if ( result != path.end() )  {
						nodeNeeded++;
						
						double nodeSuccessSubsetRuleTwo = 0.0;
						int neighborsSubsetRuleTwoApplies = 0;
						bool hasNodeStrictLFA = false;
						
						//get all other neighbors (failed neighbor will be exluded inside)
						for (vector<Node*>::iterator myiter = nbrsList.begin(); myiter != nbrsList.end(); ++myiter) { 
							Node * mynbr = *myiter;

							//avoid checking for alternate path using the assumed failed neighbouring router
							if ( mynbr->getId() == nbr->getId() ) continue; 
							
							//LFAs
							if ( mynbr->getCostToDst(dstId) <  src->getCostToDst(dstId) ) {
														
								//does this LFA provide a loop free path?
								//check if new neigbour contains the failed node in its path
								graphp->recordSPPerPair(mynbr, dstId);
								vector<int> mypath = mynbr->routingTable[dstId].shortest.nids;
								vector<int>::iterator myresult = std::find(mypath.begin(), mypath.end(), nbr->getId() );
								
								if ( myresult == mypath.end()) {
									hasNodeStrictLFA= true;
								}
							}
							
							if (!hasNodeStrictLFA) {
								neighborsSubsetRuleTwoApplies++;
								int prevNbrsNeeded = 0;
								int prevNbrsSuccess = 0;
							
								for (vector<Node*>::iterator previter = nbrsList.begin(); previter != nbrsList.end(); ++previter) { 
									Node * prevnbr = *previter;
									if ( prevnbr->getId() == nbr->getId() ||     //prev neighbor = failed neighbor
										 prevnbr->getNextHopToDst(dstId) != srcId ||  //prev neighbor does not have SRC as the next hop to DST
										 prevnbr->getId() == mynbr->getId()) continue; //prev neighbor = recovery neighbor
						
									prevNbrsNeeded++;
									
									//rule 2 condition (half checked. the other half checked above)
									if ( mynbr->getCostToDst(dstId) < prevnbr->getCostToDst(dstId) )  {
								
										//does this LFA provide a loop free path?
										//check if new neigbour contains the failed node in its path
										graphp->recordSPPerPair(mynbr, dstId);
										vector<int> mypath = mynbr->routingTable[dstId].shortest.nids;
										vector<int>::iterator myresult = std::find(mypath.begin(), mypath.end(), nbr->getId() );
										
										if ( myresult == mypath.end()) {
											prevNbrsSuccess++;
										}
										
								
									}
								
									//~ cout << " For Dst \t" << dstId << 
										//~ " with prevNeighb \t" <<  prevnbr->getId() << 
										//~ " to avoid NODE \t" << nbr->getId() <<
										//~ "\t using neighbor \t" << mynbr->getId() <<
										//~ "\t cost(n[i+1]) \t" << mynbr->getCostToDst(dstId) <<
										//~ "\t cost(n[i]) \t" << src->getCostToDst(dstId) << endl;
								}
								if (prevNbrsNeeded) {
									nodeSuccessSubsetRuleTwo += (double) prevNbrsSuccess/ prevNbrsNeeded ;
								}		
							}	
						}
						
						if (hasNodeStrictLFA) 
							nodeSuccess++;
						else if (nodeSuccessSubsetRuleTwo!=0 && neighborsSubsetRuleTwoApplies!=0)
								nodeSuccess += nodeSuccessSubsetRuleTwo / neighborsSubsetRuleTwoApplies;
					
					}
				}
			}
	    
		}
	}
	
	//cout << "# Graph connectivity" << endl ; 
	//graphp->outputDegree(cout);
	

	cout << endl <<"#Rule2 for link failures : " << linkSuccess << " out of " << linkNeeded << 
		  "  Percentage : "  <<  (double) linkSuccess / linkNeeded << endl;
	
	cout << endl << endl << endl ;

	cout << "#Rule2 for node failures : " << nodeSuccess << " out of " << nodeNeeded << 
		  "  Percentage : "  <<  (double) nodeSuccess / nodeNeeded << endl;
	
}

void rule3(Graph * graphp){
	
}

void outputCDF(map<int,int> & pdf, ostream & os) {
	
	double total = 0;
	for (map<int,int>::iterator iter = pdf.begin(); iter != pdf.end(); ++iter) {
		total += iter->second;
	}

	double half = total / 2;
	double sum = 0;
	double average = 0;
	int median = -1;
	os << "# number | pdf | cdf | percentage" << endl ;
	for (map<int, int>::iterator iter = pdf.begin(); iter != pdf.end(); ++iter) {
		sum += iter->second;
		average += iter->first * iter->second;

		if  (sum > half && median == -1) {
			median = iter->first;
		}
		os << iter->first << "\t" << iter->second << "\t" << sum << "\t" << sum / total << endl;
	}
	average /= sum;
	os << "# average " << average << endl;
	os << "# median  " << median << endl << endl;
}

bool isLinkInPath(int a, int b, vector<int> path){
	bool preva = false ;
	for (vector<int>::iterator iter = path.begin(); iter != path.end(); ++iter) {
		if (preva && (*iter == b)) return true;
		if (*iter == a) preva = true;
		else preva = false;
	}
	return false;
}

void printUsage(char* progname)
{
  cerr << progname << endl << "options";
  cerr << "\t -i <input>" << endl;
  cerr << "\t -m <ipfrr mechanism>" << endl;
}
