#ifndef HEAP_H	/* Maintain a heap of indices into an existing array */
#define HEAP_H

#include <iostream>
#include <fstream>
#include <cstdlib>

using namespace std;

void PQinit(int size, int (*)(int, int));
int  PQempty();
void PQinsert(int value);
int  PQdelmin();
int  PQreverse(int value);
void PQchange(int id);
void PQdelall();

struct IHP {
    int *pq,
        *qp;
    int  max,
         N;
} ihp;

int (*heap_cmp) (int, int);

int PQreverse(int value) {
  return ihp.qp[value];
}

void _fixUp(int hole) {     /* hole is idx of the element being moved up */
    int x = ihp.pq[hole];
    
    for ( ; hole > 1 && heap_cmp(x, ihp.pq[hole >> 1]); hole >>= 1) {
	ihp.pq[hole] = ihp.pq[hole >> 1];
	ihp.qp[ihp.pq[hole]] = hole;
    }
    ihp.pq[hole] = x;
    ihp.qp[x] = hole ;
}

void _fixDown(int hole) {
    int child;
    /* assume the order ihp.qp[ihp.pq[k]] = ihp.pq[ihp.qp[k]] = k is
       maintained before the invocation of this function */  
    int x = ihp.pq[hole];
    
    for ( ; hole * 2 <= ihp.N; hole = child) {
	child = hole * 2 ;
	if (child != ihp.N && heap_cmp(ihp.pq[child+1], ihp.pq[child]))
	    child++;
	if (heap_cmp(ihp.pq[child], x)) {
	    ihp.pq[hole] = ihp.pq[child];
	    ihp.qp[ihp.pq[child]] = hole;
	} else
	    break; /* find the right hole */
    }
    ihp.pq[hole] = x;
    ihp.qp[x] = hole;
}				       

void PQinit(int size, int (*cmp)(int, int)) {
    if (   !(ihp.pq = (int*) malloc((size+1) * sizeof (int)))
	|| !(ihp.qp = (int*) malloc((size+1) * sizeof (int)))) {
        cerr << "insufficent memory in heap\n";
		exit(1);
    }
    ihp.N = 0;
    ihp.max = size;
    heap_cmp = cmp;
}

void PQfree() {
    free(ihp.pq);
    free(ihp.qp);
}

int PQempty() { return ihp.N == 0; }
void PQdelall() { ihp.N = 0; }

int PQfull() { return ihp.N == ihp.max; }

void PQinsert(int x) {
    if (PQfull()) {
        cerr << "heap is full while inserting\n";
	exit(1);
    }
    ihp.pq[++ihp.N] = x;
    ihp.qp[x] = ihp.N;
    _fixUp(ihp.N);
}

int PQdelmin() {
    int ret;
    
    if (PQempty()) {
        cerr << "heap is empty while deleting\n";
	exit(1);
    }
    ret = ihp.pq[1];
    ihp.pq[1] = ihp.pq[ihp.N--];
    ihp.qp[ihp.pq[1]] = 1;
    _fixDown(1);
    return ret;
}

void PQchange(int k) {	/* the kth element in the array is changed */
    int pos = ihp.qp[k];
    _fixUp(pos);
    _fixDown(pos);
}
#endif /* HEAP_H */
